/*---------------------------------------------------------------------*/
/* FILE NAME  :  HPKISignSampleCrypto.c                 */
/* VERSION    :  1.0                           */
/* DATE      :  2020/10/29                       */
/*---------------------------------------------------------------------*/


/*=====================================================================*/
/*                             INCLUDE                                 */
/*=====================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>
#include <Windows.h>
#include <wincrypt.h>

/*=====================================================================*/
/*                             DEFINITION                              */
/*=====================================================================*/

#define APROVNAME TEXT("HPKI Crypto Service Provider for Authentication")
#define SPROVNAME TEXT("HPKI Crypto Service Provider for Non Repudiation")
#define HASHSIZE_SHA1 (160 / 8)

/*=====================================================================*/
/*              DEFINITIOIN OF PRIVATE FUNCTION               */
/*=====================================================================*/

/* vO̎s`@*/
void PrintUsage() {
  printf("Usage : HPKISignSample <prov_type> <method> <pin>\n");
  printf("prov_type:@auth | sign\n");
  printf("method:@getcert | sign\n");
  printf("pin:@HPKICard PIN\n");
}

/* G[̏ */
void printErrorAndExit(const char *f)
{
  DWORD errorcode;
  char errmsg[256];

  errorcode = GetLastError();
  FormatMessageA(
    FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS |
    FORMAT_MESSAGE_MAX_WIDTH_MASK,
    NULL,
    errorcode,
    MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT),
    errmsg, sizeof(errmsg) / sizeof(TCHAR),
    NULL);
  printf("%s: 0x%08x (%s).\n", f, errorcode, errmsg);

  exit(EXIT_FAILURE);
}

/* ef[^(ؖf[^Af[^)16i\ŕWo */
void printHex(const BYTE* data, DWORD len, BOOL limit)
{
  DWORD displen;

  if (limit && len > 256)
    displen = 256;
  else
    displen = len;
  for (DWORD i = 0; i < displen; i++) {
    printf("%02x", *(data + i));
    if (i % 16 == 15 || i == displen - 1)
      printf("\n");
    else
      printf(" ");
  }
  if (displen < len)
    printf("...\n");
}

/*
 * getCert: pҏؖ擾
 * ȉ̎菇ŏؖ擾B
 * 1. voC_nh擾
 * 2. RSAyÃnh擾
 * 3. ؖf[^擾
 * 4. nhj
 * 5. voC_nh
 */

void getCert(LPCTSTR pProvName)
{
  HCRYPTPROV hProv;
  HCRYPTKEY  hKey;
  BYTE* certData;
  DWORD certLen;

  /* (1) ÍvoC_ƃReiw肵āAvoC_nh擾B */
  if (!CryptAcquireContext(&hProv, NULL, pProvName, PROV_RSA_FULL, 0)) {
    printErrorAndExit("CryptAcquireContext");
  }

  /* (2) Rei̕ێRSAyÃnh擾B */
  if (!CryptGetUserKey(hProv, AT_SIGNATURE, &hKey)) {
    printErrorAndExit("CryptGetUserKey");
  }

  /* (3) nhɕt闘pҏؖf[^̃f[^擾B */
  certLen = 0;
  if (!CryptGetKeyParam(hKey, KP_CERTIFICATE, NULL, &certLen, 0)) {
    printErrorAndExit("CryptGetKeyParam(1)");
  }

  certData = malloc(certLen);
  if (certData == NULL) {
    printf("Not enough memory.\n");
    exit(EXIT_FAILURE);
  }

  /* (4) nhɕt闘pҏؖf[^̃f[^擾B */
  if (!CryptGetKeyParam(hKey, KP_CERTIFICATE, certData, &certLen, 0)) {
    printErrorAndExit("CryptGetKeyParam(2)");
  }

  printf("Certificate:\n");
  printHex(certData, certLen, FALSE);

  free(certData);

  /* (5) nhjB */
  if (!CryptDestroyKey(hKey)) {
    printErrorAndExit("CryptDestroyKey");
  }

  /* (6) voC_nhB */
  if (!CryptReleaseContext(hProv, 0)) {
    printErrorAndExit("CryptReleaseContext");
  }
}

/*
 * sign: 
 * ȉ̎菇ŏ{B
 *  1. voC_nh擾
 *  2. RSAyÃnh擾
 *  3. ؖf[^擾
 *  4. nbVIuWFNg
 *  5. nbVlݒ
 *  6. f[^擾
 *  7. PINݒ
 *  8. f[^擾
 *  9. nbVIuWFNgj
 * 10. nhj
 * 11. voC_nh
 */

void sign(LPCTSTR pProvName, const BYTE *pPinData, DWORD pinLen)
{
  HCRYPTPROV hProv;
  HCRYPTKEY  hKey;
  HCRYPTHASH hHash;
  BYTE* certData;
  DWORD certLen;
  BYTE hashData[256];
  BYTE* sigData;
  DWORD sigLen;

  /* (7) ÍvoC_ƃReiw肵āAvoC_nh擾B */
  if (!CryptAcquireContext(&hProv, NULL, pProvName, PROV_RSA_FULL, 0)) {
    printErrorAndExit("CryptAcquireContext");
  }

  /* (8) Rei̕ێRSAyÃnh擾B */
  if (!CryptGetUserKey(hProv, AT_SIGNATURE, &hKey)) {
    printErrorAndExit("CryptGetUserKey");
  }

  /* (9) nhɕt闘pҏؖf[^̃f[^擾B */
  certLen = 0;
  if (!CryptGetKeyParam(hKey, KP_CERTIFICATE, NULL, &certLen, 0)) {
    printErrorAndExit("CryptGetKeyParam(1)");
  }

  certData = malloc(certLen);
  if (certData == NULL) {
    printf("Not enough memory.\n");
    exit(EXIT_FAILURE);
  }

  /* (10) nhɕt闘pҏؖf[^̃f[^擾B */
  if (!CryptGetKeyParam(hKey, KP_CERTIFICATE, certData, &certLen, 0)) {
    printErrorAndExit("CryptGetKeyParam(2)");
  }

  printf("Certificate:\n");
  printHex(certData, certLen, TRUE);

  free(certData);

  /* (11) nbVIuWFNg̐ */
  if (!CryptCreateHash(hProv, CALG_SHA1, 0, 0, &hHash)) {
    printErrorAndExit("CryptCreateHash");
  }

  /* (12) Ώۃf[^̃nbVl𐶐 ({Tvł̓_~[f[^gp) */
  memset(hashData, 0xa5, HASHSIZE_SHA1);

  /* (13) nbVIuWFNg̃p[^ݒ */
  if (!CryptSetHashParam(hHash, HP_HASHVAL, hashData, 0)) {
    printErrorAndExit("CryptSetHashParam");
  }

  /* (14) nbVlɏsʂ̃f[^擾B */
  sigLen = 0;
  if (!CryptSignHash(hHash, AT_SIGNATURE, NULL, 0, NULL, &sigLen)) {
    printErrorAndExit("CryptSignHash(1)");
  }

  /* (15) pPINݒ肷B */
  if (!CryptSetProvParam(hProv, PP_SIGNATURE_PIN, pPinData, pinLen)) {
    printErrorAndExit("CryptSetProvParam");
  }

  sigData = malloc(sigLen);
  if (sigData == NULL) {
    printf("Not enough memory.\n");
    exit(EXIT_FAILURE);
  }

  /* (16) nbVlɏsʂ擾B */
  if (!CryptSignHash(hHash, AT_SIGNATURE, NULL, 0, sigData, &sigLen)) {
    printErrorAndExit("CryptSignHash(2)");
  }

  printf("Signature:\n");
  printHex(sigData, sigLen, FALSE);

  free(sigData);

  /* (17) nbVIuWFNgj */
  if (!CryptDestroyHash(hHash)) {
    printErrorAndExit("CryptDestroyHash");
  }

  /* (18) nhjB */
  if (!CryptDestroyKey(hKey)) {
    printErrorAndExit("CryptDestroyKey");
  }

  /* (19) voC_nhB */
  if (!CryptReleaseContext(hProv, 0)) {
    printErrorAndExit("CryptReleaseContext");
  }
}


/*=====================================================================*/
/*                    Start of main()                             */
/*=====================================================================*/

int main(int argc, char *argv[])
{
  LPCTSTR pProvName;
  BYTE* pPin;
  DWORD pinLen;

  /*--------------------*/
  /* (A) ̃`FbN */
  /*--------------------*/

  if (argc != 4)
  {
    printf("s`܂B\n");
    PrintUsage();
    exit(EXIT_SUCCESS);
  }

  /* CSPmF
   * auth : dqFؗpCSP
   * sign : dqpCSP
  */
  switch (*argv[1]) {
  case 'a':
  case 'A':
    pProvName = APROVNAME;
    break;
  case 's':
  case 'S':
    pProvName = SPROVNAME;
    break;
  default:
    printf("sp[^<prov_type>܂B\n");
    PrintUsage();
    exit(EXIT_FAILURE);
    break;
  }

  /* PIN */
  pPin = argv[3];
  pinLen = strlen(argv[3]);

  /*--------------------*/
  /* (B) ̎s    */
  /*--------------------*/

  switch (*argv[2]) {
    /* ؖ擾 */
  case 'g':
  case 'G':
    printf("ؖ擾 Jn\n");
    getCert(pProvName);
    printf("\nؖ擾 \n");
    break;
    /*  */
  case 's':
  case 'S':
    printf(" Jn\n");
    sign(pProvName, pPin, pinLen);
    printf("\n \n");
    break;
  default:
    printf("sp[^<method>܂B\n");
    PrintUsage();
    exit(EXIT_FAILURE);
    break;
  }

  exit(EXIT_SUCCESS);
}
/*=====================================================================*/
/*             End of main()                   */
/*=====================================================================*/